import pytest

from compman import soaringspot


@pytest.mark.asyncio
async def test_fetch_competitions() -> None:
    # WHEN
    comps = await soaringspot.fetch_competitions()

    # THEN
    # list of competitions changes all the time, so all we can assert is that
    # list is not empty.

    assert len(comps) > 0


@pytest.mark.asyncio
async def test_fetch_downloads() -> None:
    # GIVEN
    ssurl = "https://www.soaringspot.com/en_gb/wgc2018pl/"

    # WHEN
    downloads = await soaringspot.fetch_downloads(ssurl)

    # THEN
    assert len(downloads) == 2

    dl0 = downloads[0]
    assert dl0.filename == "wgc2018_airspace_v5.1_openair.txt"
    assert dl0.kind == soaringspot.DownloadableFileType.AIRSPACE

    dl1 = downloads[1]
    assert dl1.filename == "pz_wgc_2018_v1.2.cup"
    assert dl1.kind == soaringspot.DownloadableFileType.WAYPOINT


@pytest.mark.asyncio
async def test_fetch_file() -> None:
    # GIVEN
    ssurl = "https://www.soaringspot.com/en_gb/wgc2018pl/"
    downloads = await soaringspot.fetch_downloads(ssurl)
    dl0 = downloads[0]

    # WHEN
    fetched = await soaringspot.fetch_file(dl0.href)
    content = fetched.read()

    # THEN
    assert len(content) == 138067



from typing import Optional, Dict, Any, List, IO
import logging
import os
import json
from dataclasses import dataclass, field

DEFAULT_DATADIR = "~/.compman"

_DATADIR: Optional[str]
_SETTINGS: Optional["Settings"] = None

log = logging.getLogger("compman")


@dataclass
class Settings:
    current_competition_id: Optional[str] = None

    @classmethod
    def fromdict(cls, data: Dict[str, Any]) -> "Settings":
        return cls(current_competition_id=data.get("current_competition_id"))

    def asdict(self) -> Dict[str, Any]:
        return {"current_competition_id": self.current_competition_id}


@dataclass
class StoredCompetition:
    id: str
    title: str
    soaringspot_url: Optional[str] = None
    airspace: Optional[str] = None
    waypoints: Optional[str] = None
    profiles: List[str] = field(default_factory=list)

    @classmethod
    def fromdict(cls, id: str, data: Dict[str, Any]) -> "StoredCompetition":
        return cls(
            id=id,
            title=data["title"],
            soaringspot_url=data.get("soaringspot_url"),
            airspace=data.get("airspace"),
            waypoints=data.get("waypoints"),
            profiles=data.get("profiles") or [],
        )

    def asdict(self) -> Dict[str, Any]:
        return {
            "title": self.title,
            "soaringspot_url": self.soaringspot_url,
            "airspace": self.airspace,
            "waypoints": self.waypoints,
            "profiles": self.profiles,
            "version": 1,
        }

    def add_profile(self, profile) -> None:
        if profile not in self.profiles:
            self.profiles.append(profile)

    def remove_profile(self, profile: str) -> None:
        if profile in self.profiles:
            self.profiles.remove(profile)


@dataclass
class StoredFile:
    name: str
    size: Optional[int]

    def format_size(self):
        if self.size is None:
            return "?"
        suffix = "B"
        for unit in ["", "Ki", "Mi", "Gi"]:
            if abs(self.size) < 1024.0:
                return "%3.1f%s%s" % (self.size, unit, suffix)
            self.size /= 1024.0
        return "%.1f%s%s" % (self.size, "Ti", suffix)


def init(datadir: str) -> None:
    global _DATADIR, _SETTINGS
    _DATADIR = datadir
    os.makedirs(datadir, mode=0o755, exist_ok=True)
    _SETTINGS = None


def deinit() -> None:
    global _DATADIR, _SETTINGS
    _DATADIR = None
    _SETTINGS = None


def get_settings() -> Settings:
    global _SETTINGS
    if _SETTINGS is None:
        _SETTINGS = load_settings()

    return _SETTINGS


def load_settings() -> Settings:
    fname = _get_settings_fname()
    if os.path.exists(fname):
        with open(fname) as f:
            try:
                data = json.load(f)
            except json.decoder.JSONDecodeError:
                log.error(f"Cannot load settings file: {fname}")
                data = {}
    else:
        data = {}
    return Settings.fromdict(data)


def save_settings() -> None:
    settings = get_settings()
    fname = _get_settings_fname()
    with open(fname, "w") as f:
        json.dump(settings.asdict(), f, indent=2)


def save_competition(comp: StoredCompetition, set_current: bool = False) -> None:
    compdir = _get_compdir(comp.id)

    if not os.path.exists(compdir):
        os.mkdir(compdir, 0o755)

    with open(_get_compconfigname(comp.id), "wt") as f:
        compdict = comp.asdict()
        json.dump(compdict, f, indent=2)
        f.write("\n")

    if set_current:
        get_settings().current_competition_id = comp.id
        save_settings()


def load_competition(cid: str) -> Optional[StoredCompetition]:
    if not exists(cid):
        return None

    with open(_get_compconfigname(cid), "rt") as f:
        compdict = json.load(f)
        comp = StoredCompetition.fromdict(cid, compdict)

    return comp


def list_competitions() -> List[StoredCompetition]:
    competitions = []
    for cid in os.listdir(_DATADIR):
        comp = load_competition(cid)
        if comp is None:
            continue
        competitions.append(comp)

    return sorted(competitions, key=lambda c: c.title)


def store_file(cid: str, filename: str, contents: IO[bytes]) -> StoredFile:
    cdir = _get_compdir(cid)
    fullname = os.path.join(cdir, filename)
    with open(fullname, "wb") as f:
        f.write(contents.read())
    return StoredFile(name=filename, size=os.path.getsize(fullname))


def get_airspace_files(cid: str) -> List[StoredFile]:
    return _get_files(cid, ".txt")


def get_waypoint_files(cid: str) -> List[StoredFile]:
    return _get_files(cid, ".cup")


def _get_files(cid: str, ext: str) -> List[StoredFile]:
    cdir = _get_compdir(cid)

    files = []
    for entry in os.scandir(cdir):
        if entry.is_dir():
            continue
        if not entry.name.endswith(ext):
            continue
        stat = entry.stat()
        files.append(StoredFile(name=entry.name, size=stat.st_size))
    return files


def get_full_file_path(cid: str, fname: str) -> str:
    return os.path.abspath(os.path.join(_get_compdir(cid), fname.strip()))


def exists(cid: str) -> bool:
    configfname = _get_compconfigname(cid)
    return os.path.exists(configfname)


def _get_compdir(cid: str) -> str:
    assert _DATADIR
    return os.path.join(_DATADIR, cid)


def _get_compconfigname(cid: str) -> str:
    compdir = _get_compdir(cid)
    return os.path.join(compdir, "competition.json")


def _get_settings_fname() -> str:
    assert _DATADIR
    return os.path.join(_DATADIR, "settings.json")



import asyncio

import pytest

from compman.ui.soaringspot import SoaringSpotPickerScreen
from compman.soaringspot import SoaringSpotContest
from compman import storage

from tests.fixtures.soaringspot import SoaringSpotFixture
from tests.fixtures.activitytestbed import ActivityTestbed


@pytest.mark.asyncio
async def test_soaringspot_no_contests(
    soaringspot: SoaringSpotFixture, activity_testbed: ActivityTestbed
) -> None:
    async with activity_testbed.shown(SoaringSpotPickerScreen):
        contents = activity_testbed.render()
        assert "Pick a new competition" in contents


@pytest.mark.asyncio
async def test_soaringspot_view(
    soaringspot: SoaringSpotFixture, activity_testbed: ActivityTestbed
) -> None:

    soaringspot.competitions = [
        SoaringSpotContest(
            id="one",
            href="http://soaringspot.local/one",
            title="One",
            description="First contest",
        ),
        SoaringSpotContest(
            id="two",
            href="http://soaringspot.local/two",
            title="Two",
            description="Second contest",
        ),
    ]

    async with activity_testbed.shown(SoaringSpotPickerScreen):
        contents = activity_testbed.render()
        assert "Pick a new competition" in contents

        await activity_testbed.gather_tasks()
        contents = activity_testbed.render()
        assert "One" in contents
        assert "Two" in contents

        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "One"


@pytest.mark.asyncio
async def test_soaringspot_select(
    storage_dir: str, soaringspot: SoaringSpotFixture, activity_testbed: ActivityTestbed
) -> None:
    # GIVEN
    soaringspot.competitions = [
        SoaringSpotContest(
            id="one",
            href="http://soaringspot.local/one",
            title="One",
            description="First contest",
        ),
        SoaringSpotContest(
            id="two",
            href="http://soaringspot.local/two",
            title="Two",
            description="Second contest",
        ),
    ]

    # WHEN
    async with activity_testbed.shown(SoaringSpotPickerScreen):
        await activity_testbed.gather_tasks()
        await activity_testbed.keypress("down")
        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "Two"
        assert "Second contest" in activity_testbed.render()

        await activity_testbed.keypress("enter")

        # THEN
        res = await activity_testbed.response()
        assert isinstance(res, storage.StoredCompetition)
        assert res.id == "two"
        assert res.title == "Two"
        assert res.soaringspot_url == "http://soaringspot.local/two"


@pytest.mark.asyncio
async def test_soaringspot_existing(
    storage_dir: str, soaringspot: SoaringSpotFixture, activity_testbed: ActivityTestbed
) -> None:
    # GIVEN
    storage.save_competition(storage.StoredCompetition(id="one", title="One"))

    soaringspot.competitions = [
        SoaringSpotContest(
            id="one",
            href="http://soaringspot.local/one",
            title="One",
            description="First contest",
        ),
        SoaringSpotContest(
            id="two",
            href="http://soaringspot.local/two",
            title="Two",
            description="Second contest",
        ),
    ]

    # WHEN
    async with activity_testbed.shown(SoaringSpotPickerScreen):
        await activity_testbed.gather_tasks()
        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "One"
        assert "First contest" in activity_testbed.render()
        await activity_testbed.keypress("enter")

        # THEN
        res = await activity_testbed.response()
        assert isinstance(res, storage.StoredCompetition)
        assert res.id == "one"
        assert res.title == "One"
        assert res.soaringspot_url == "http://soaringspot.local/one"



from typing import Optional, List
from shutil import copyfile
import os


XCSOAR_DIR: Optional[str] = None


class XCSoarProfile:
    def __init__(self, filename):
        self.filename = filename
        self.xcsoardir = os.path.dirname(filename)

        with open(filename, "rt") as f:
            self.lines = f.readlines()

    def save(self):
        with open(self.filename, "wt") as f:
            f.write("".join(self.lines))

    def set_airspace(self, filename: str) -> None:
        fname = "compman-airspace.txt"
        copyfile(filename, os.path.join(self.xcsoardir, fname))
        self._set_option("AirspaceFile", f"%LOCAL_PATH%\\{fname}")

    def set_waypoint(self, filename: str) -> None:
        fname = "compman-waypoints.cup"
        copyfile(filename, os.path.join(self.xcsoardir, fname))
        self._set_option("WPFile", f"%LOCAL_PATH%\\{fname}")

    def _set_option(self, key: str, value: str) -> None:
        modified_line = f'{key}="{value}"\n'
        for n, line in enumerate(self.lines):
            k, v = line.split("=", maxsplit=1)
            if k == key:
                self.lines[n] = modified_line
                break
        else:
            self.lines.append(modified_line)


def init(xcsoar_dir: str = None) -> None:
    global XCSOAR_DIR
    if xcsoar_dir is None:
        xcsoar_dir = find_xcsoar_dir()
    XCSOAR_DIR = xcsoar_dir


def deinit() -> None:
    global XCSOAR_DIR
    XCSOAR_DIR = None


def find_xcsoar_dir() -> str:
    home = os.path.expanduser("~")
    xcsoardir = os.path.join(home, ".xcsoar")
    if not os.path.exists(xcsoardir):
        raise FileNotFoundError(xcsoardir)
    return xcsoardir


def list_xcsoar_profiles() -> List[str]:
    assert XCSOAR_DIR is not None

    profiles = []
    for fname in os.listdir(XCSOAR_DIR):
        if fname.endswith(".prf"):
            profiles.append(fname)

    return profiles


def get_xcsoar_profile_filename(profile_fname: str) -> str:
    assert XCSOAR_DIR is not None
    return os.path.join(XCSOAR_DIR, profile_fname)


def get_xcsoar_profile(profile_fname: str) -> XCSoarProfile:
    profile_fullname = get_xcsoar_profile_filename(profile_fname)
    return XCSoarProfile(profile_fullname)



import logging

import urwid

from compman.ui import widget
from compman.ui.activity import Activity
from compman import storage


log = logging.getLogger("compman")


class SelectCompetitionScreen(Activity):
    def create_view(self):
        self._items = []
        self._populate_stored_comps()
        picker = urwid.Pile(self._items)

        add_btn = widget.CMButton("New competition")
        self.connect_async(add_btn, "click", self._on_new_competition)
        cancel_btn = widget.CMButton("Cancel")
        urwid.connect_signal(cancel_btn, "click", self._on_cancel)

        buttons = urwid.GridFlow([add_btn, cancel_btn], 22, 2, 1, "left")

        mainview = urwid.Pile([picker, urwid.Divider(), buttons])

        return urwid.Frame(
            urwid.Filler(widget.CMScreenPadding(mainview), valign=urwid.TOP),
            header=widget.CMScreenHeader("Switch competition"),
        )

    def _populate_stored_comps(self):
        for comp in storage.list_competitions():
            btn = widget.CMSelectableListItem(comp.title)
            urwid.connect_signal(btn, "click", self._on_competition_selected, comp)
            self._items.append(btn)

    def _on_competition_selected(self, btn, comp):
        self._comp_selected(comp)

    async def _on_new_competition(self):
        from compman.ui.soaringspot import SoaringSpotPickerScreen

        screen = SoaringSpotPickerScreen(self.container)
        res = await self.run_activity(screen)
        if res is not None:
            self._comp_selected(res)

    def _comp_selected(self, comp):
        storage.get_settings().current_competition_id = comp.id
        storage.save_settings()

        self.finish(comp)

        from compman.ui.compdetails import CompetitionDetailsScreen

        details = CompetitionDetailsScreen(self.container)
        details.show()

    def _on_cancel(self, btn):
        self.finish(None)



import os
import io


from compman import storage


def test_get_settings_new(storage_dir) -> None:
    settings = storage.get_settings()
    assert settings is not None
    assert settings.current_competition_id is None


def test_get_settings_bad_file(storage_dir) -> None:
    # GIVEN
    settings_fname = storage._get_settings_fname()
    with open(settings_fname, "w") as f:
        f.write(";garbage;")

    # WHEN
    settings = storage.get_settings()

    # THEN
    assert settings is not None
    assert settings.current_competition_id is None


def test_save_settings_new(storage_dir) -> None:
    # WHEN
    storage.save_settings()

    # THEN
    settings_fname = storage._get_settings_fname()
    assert os.path.exists(settings_fname)


def test_save_settings_modified(storage_dir) -> None:
    settings = storage.get_settings()
    settings.current_competition_id = "hello"

    # WHEN
    storage.save_settings()

    # THEN
    settings = storage.load_settings()
    assert settings.current_competition_id == "hello"


def test_list_competitions_empty(storage_dir) -> None:
    # WHEN
    comps = storage.list_competitions()

    # THEN
    assert comps == []


def test_save_competition(storage_dir) -> None:
    # GIVEN
    comp = storage.StoredCompetition(
        id="first",
        title="First Competition",
        soaringspot_url="http://localhost",
        airspace="airspace.txt",
        waypoints="waypoints.txt",
    )

    # WHEN
    storage.save_competition(comp)

    # THEN
    comps = storage.list_competitions()
    assert [c.id for c in comps] == ["first"]


def test_load_competition_missing(storage_dir) -> None:
    # WHEN
    comp = storage.load_competition("missing")

    # THEN
    assert comp is None


def test_load_competition_existing(storage_dir) -> None:
    # GIVEN
    comp = storage.StoredCompetition(
        id="first",
        title="First Competition",
        soaringspot_url="http://localhost",
        airspace="airspace.txt",
        waypoints=None,
    )
    storage.save_competition(comp)

    # WHEN
    loaded = storage.load_competition("first")
    assert loaded is not None

    # THEN
    assert loaded.id == "first"
    assert loaded.title == "First Competition"
    assert loaded.soaringspot_url == "http://localhost"
    assert loaded.airspace == "airspace.txt"
    assert loaded.waypoints == None


def test_store_file(storage_dir) -> None:
    # GIVEN
    comp = storage.StoredCompetition(id="first", title="First Competition")
    storage.save_competition(comp)

    # WHEN
    storage.store_file("first", "airspace.txt", io.BytesIO(b"test"))

    # THEN
    files = storage.get_airspace_files("first")
    assert [f.name for f in files] == ["airspace.txt"]
    files = storage.get_waypoint_files("first")
    assert files == []


def test_get_waypoint_files(storage_dir) -> None:
    # GIVEN
    comp = storage.StoredCompetition(id="first", title="First Competition")
    storage.save_competition(comp)

    # WHEN
    storage.store_file("first", "waypoint.cup", io.BytesIO(b"test"))

    files = storage.get_waypoint_files("first")
    assert [f.name for f in files] == ["waypoint.cup"]
    files = storage.get_airspace_files("first")
    assert files == []


def test_storedfile_format_size() -> None:
    assert storage.StoredFile("", size=1024).format_size() == "1.0KiB"
    assert storage.StoredFile("", size=1).format_size() == "1.0B"
    assert storage.StoredFile("", size=100000).format_size() == "97.7KiB"
    assert storage.StoredFile("", size=10000000).format_size() == "9.5MiB"
    assert storage.StoredFile("", size=None).format_size() == "?"
    assert storage.StoredFile("", size=2 ** 32).format_size() == "4.0GiB"
    assert storage.StoredFile("", size=2 ** 48).format_size() == "256.0TiB"


def test_get_full_file_path(storage_dir) -> None:
    path = storage.get_full_file_path("test", "airspace.txt")
    assert path == os.path.join(storage_dir, "test", "airspace.txt")



import os
import shutil
import asyncio

import pytest

from compman import storage
from compman import xcsoar
from .fixtures.soaringspot import SoaringSpotFixture
from .fixtures.activitytestbed import ActivityTestbed

HERE = os.path.dirname(__file__)


@pytest.fixture()
def soaringspot():
    ssf = SoaringSpotFixture()
    ssf.setUp()
    try:
        yield ssf
    finally:
        ssf.tearDown()


@pytest.fixture
def storage_dir(tmpdir):
    storage.init(tmpdir)
    yield tmpdir
    storage.deinit()


@pytest.fixture
def activity_testbed(mocker):
    yield ActivityTestbed(mocker)


@pytest.fixture
def xcsoar_dir(tmpdir):
    xcsoar_sample_dir = os.path.join(HERE, "fixtures", "xcsoar")
    xcsoardir = os.path.join(tmpdir, ".xcsoar")
    shutil.copytree(xcsoar_sample_dir, xcsoardir)
    xcsoar.init(xcsoardir)
    yield xcsoardir
    xcsoar.deinit()


@pytest.fixture
def async_sleep(mocker):
    realsleep = asyncio.sleep

    async def sleep_mock(time: float) -> None:
        await realsleep(0)

    mocker.patch("asyncio.sleep", sleep_mock)



from typing import List, IO
from unittest import mock
import io
import asyncio


from compman.soaringspot import SoaringSpotContest, SoaringSpotDownloadableFile


class SoaringSpotFixture:
    competitons: List[SoaringSpotContest]
    files: List[SoaringSpotDownloadableFile]
    file_contents: bytes

    def setUp(self) -> None:
        self.reset()

        self._patches = [
            mock.patch(
                "compman.soaringspot.fetch_competitions", self.fetch_competitions_mock
            ),
            mock.patch(
                "compman.soaringspot.fetch_downloads", self.fetch_downloads_mock
            ),
            mock.patch("compman.soaringspot.fetch_file", self.fetch_file_mock),
        ]

        for p in self._patches:
            p.start()

    def tearDown(self) -> None:
        for p in self._patches:
            p.stop()

    def reset(self):
        self.competitions = []
        self.files = []
        self.file_contents = b""

    async def fetch_competitions_mock(self) -> List[SoaringSpotContest]:
        # Downloading...
        await asyncio.sleep(0)
        return self.competitions

    async def fetch_downloads_mock(
        self, comp_url: str
    ) -> List[SoaringSpotDownloadableFile]:
        # Downloading...
        await asyncio.sleep(0)
        return self.files

    async def fetch_file_mock(self, file_url: str) -> IO[bytes]:
        # Downloading...
        await asyncio.sleep(0)
        return io.BytesIO(self.file_contents)



from typing import Sequence
import os

import urwid
from ovshell import protocol

from compman import storage
from compman import xcsoar
from compman import main
from compman.ui.mainmenu import MainMenuScreen


def extension(id: str, shell: protocol.OpenVarioShell) -> protocol.Extension:
    return CompmanExtension(id, shell)


class CompmanExtension(protocol.Extension):
    title = "Compman"

    def __init__(self, id: str, shell: protocol.OpenVarioShell) -> None:
        self.id = id
        self.shell = shell

    def list_apps(self) -> Sequence[protocol.App]:
        return [CompmanShellApp(self.shell)]


class CompmanShellApp(protocol.App):
    name = "compman"
    title = "Compman"
    description = "Competition manager"
    priority = 70

    def __init__(self, shell: protocol.OpenVarioShell) -> None:
        self.shell = shell

    def launch(self) -> None:
        datadir = os.environ.get("COMPMAN_DATADIR", storage.DEFAULT_DATADIR)
        storage.init(datadir)
        xcsoardir = os.environ.get("COMPMAN_XCSOARDIR", None)
        xcsoar.init(xcsoardir)
        main.setup_logging(datadir)

        act = CompmanShellActivity(self.shell)
        self.shell.screen.push_activity(act, palette=main.PALETTE)


class CompmanShellActivity(protocol.Activity):
    def __init__(self, shell: protocol.OpenVarioShell) -> None:
        self.shell = shell

    def create(self) -> urwid.Widget:
        container = urwid.WidgetPlaceholder(urwid.SolidFill(" "))
        screen = MainMenuScreen(container)
        screen.on_exit(self._exit)
        screen.show()
        return container

    def _exit(self) -> None:
        self.shell.screen.pop_activity()



from typing import List, IO
import re
import io
from enum import Enum
from dataclasses import dataclass
from aiohttp import ClientSession

from lxml import etree


SOARINGSPOT_URL = "https://www.soaringspot.com"


class DownloadableFileType(Enum):
    UNKNOWN = "unknown"
    AIRSPACE = "airspace"
    WAYPOINT = "waypoint"


@dataclass
class SoaringSpotDownloadableFile:
    filename: str
    href: str
    kind: DownloadableFileType = DownloadableFileType.UNKNOWN


@dataclass
class SoaringSpotContest:
    id: str
    href: str
    title: str
    description: str


async def fetch_competitions() -> List[SoaringSpotContest]:
    async with ClientSession() as session:
        async with session.get(SOARINGSPOT_URL) as response:
            html = await response.text()

    # Parse html
    parser = etree.HTMLParser()
    root = etree.parse(io.StringIO(html), parser)
    contests = []
    for contestelem in root.xpath("//*[@class='contest']"):
        linkel = contestelem.xpath("h3//a")[0]
        title = linkel.text
        href = linkel.attrib["href"]
        descr = _extract_text(contestelem.xpath("*[@class='info']"))
        cid = href.strip(" /").split("/")[-1]

        contests.append(
            SoaringSpotContest(
                id=cid, href=f"{SOARINGSPOT_URL}{href}", title=title, description=descr
            )
        )

    return contests


async def fetch_downloads(comp_url: str) -> List[SoaringSpotDownloadableFile]:
    dl_url = f"{_sanitize_url(comp_url)}/downloads"
    async with ClientSession() as session:
        async with session.get(dl_url) as response:
            html = await response.text()

    parser = etree.HTMLParser()
    root = etree.parse(io.StringIO(html), parser)

    dls = []
    for anode in root.xpath("//ul[@class='contest-downloads']/li/a"):
        filename = " ".join(anode.itertext()).strip()
        url = anode.attrib["href"]
        if not url.startswith("http"):
            url = f"{SOARINGSPOT_URL}{url}"
        dl = SoaringSpotDownloadableFile(filename=filename, href=url)
        if filename.endswith(".txt"):
            dl.kind = DownloadableFileType.AIRSPACE
        elif filename.endswith(".cup"):
            dl.kind = DownloadableFileType.WAYPOINT

        if dl.kind != DownloadableFileType.UNKNOWN:
            dls.append(dl)

    return dls


async def fetch_file(file_url: str) -> IO[bytes]:
    async with ClientSession() as session:
        async with session.get(file_url) as response:
            content = await response.read()

    return io.BytesIO(content)


def _extract_text(els) -> str:
    texts: List[str] = []
    for el in els:
        texts.extend(el.itertext())

    joined = " ".join(texts)
    return re.sub(r"[ \n]+", " ", joined).strip()


def _sanitize_url(url: str) -> str:
    return url.rstrip("/")



import os
import sys
import asyncio
import logging
import argparse

import urwid

from compman import storage
from compman import xcsoar

log = logging.getLogger("compman")


parser = argparse.ArgumentParser(description="Competition manager for Openvario")
parser.add_argument(
    "--datadir",
    default=os.environ.get("COMPMAN_DATADIR", storage.DEFAULT_DATADIR),
    help=(
        "Path to directory where compman stores data. By default "
        "~/.compman. Also can be set with COMPMAN_DATADIR environment variable"
    ),
)
parser.add_argument(
    "--xcsoardir",
    default=os.environ.get("COMPMAN_XCSOARDIR", None),
    help=(
        "Path to xcsoar home directory. By default ~/.xcsoar. Also can be set "
        "with COMPMAN_XCSOARDIR environment variable"
    ),
)


PALETTE = [
    ("text", "white", "black", ""),
    ("btn focus", "white", "dark red", ""),
    ("btn normal", "white", "dark blue", ""),
    ("li normal", "light cyan", "black", ""),
    ("li focus", "white", "dark red", ""),
    ("pg normal", "white", "black", "standout"),
    ("pg complete", "white", "dark magenta"),
    ("pg smooth", "dark magenta", "black"),
    ("screen header", "yellow", "black", ""),
    ("bg", "light gray", "black", ""),
    ("success message", "light green", "black", ""),
    ("success banner", "white", "dark green", ""),
    ("error message", "light red", "black", ""),
    ("error banner", "white", "dark red", ""),
    ("progress", "light magenta", "black", ""),
    ("remark", "dark gray", "black", ""),
]


def exit() -> None:
    raise urwid.ExitMainLoop()


def startui(urwidloop) -> None:
    # Show the splash screen as soon as possible
    urwidloop.draw_screen()

    # We can now import and show the main menu screen
    container = urwid.WidgetPlaceholder(urwid.SolidFill(" "))
    urwidloop.widget = urwid.AttrMap(container, "bg")

    from compman.ui.mainmenu import MainMenuScreen

    screen = MainMenuScreen(container)
    screen.on_exit(exit)
    screen.show()


def debounce_esc(keys, raw):
    # For some weird reason, SteFly remote stick sends two "Escape" key presses
    # when user presses X button once. Whatever reason might be, this is a
    # permanent deision and we have to deal with that. We still want to handle
    # single escape keypresses though, to support input devices that behave
    # sanely.
    filtered = []
    escpressed = False
    for k in keys:
        if k == "esc":
            if escpressed:
                escpressed = False
                filtered.append(k)
            else:
                escpressed = True
        else:
            if escpressed:
                filtered.append("esc")
                escpressed = False
            filtered.append(k)
    if escpressed:
        filtered.append("esc")
    return filtered


def setup_logging(datadir: str) -> None:
    logfname = os.path.join(datadir, "compman.log")
    logging.basicConfig(filename=logfname, level=logging.INFO)
    log.info(f"Starting compman with data dir in '{datadir}'")


def run(argv) -> None:
    args = parser.parse_args(argv)
    datadir = os.path.expanduser(args.datadir)

    storage.init(datadir)
    xcsoar.init(args.xcsoardir)
    setup_logging(datadir)

    asyncioloop = asyncio.get_event_loop()
    asyncioloop.set_debug(True)
    evl = urwid.AsyncioEventLoop(loop=asyncioloop)

    btxt = urwid.BigText("Compman", urwid.font.Thin6x6Font())
    splash = urwid.Filler(urwid.Padding(btxt, "center", "clip"), "middle")
    urwidloop = urwid.MainLoop(
        splash, palette=PALETTE, event_loop=evl, input_filter=debounce_esc
    )

    asyncioloop.call_soon(startui, urwidloop)
    try:
        urwidloop.run()
        log.info("Exiting normally")
    except KeyboardInterrupt:
        log.info("Killed")


def main():
    run(sys.argv[1:])



from typing import Callable, Optional
import urwid

import compman
from compman import storage
from compman.ui import widget
from compman.ui.activity import Activity


class MainMenuScreen(Activity):
    exithandler: Optional[Callable[[], None]] = None

    def show(self) -> None:
        super().show()

        if storage.get_settings().current_competition_id is None:
            self._run_screen("welcome")
        else:
            self._run_screen("details")

    def finish(self, result) -> None:
        super().finish(result)
        if self.exithandler is not None:
            self.exithandler()

    def on_exit(self, exithandler: Callable[[], None]) -> None:
        self.exithandler = exithandler

    def create_view(self):
        btxt = urwid.BigText("Compman", urwid.font.Thin6x6Font())
        hpad = urwid.Padding(urwid.AttrMap(btxt, "screen header"), urwid.CENTER, "clip")

        m_select_comp = widget.CMSelectableListItem("Select Competition")
        urwid.connect_signal(
            m_select_comp, "click", self._run_screen, user_args=["select"]
        )

        m_details = widget.CMSelectableListItem("Current Competition")
        urwid.connect_signal(
            m_details, "click", self._run_screen, user_args=["details"]
        )
        m_exit = widget.CMSelectableListItem("Exit")
        urwid.connect_signal(m_exit, "click", self._on_exit)

        menu = urwid.Pile([m_details, m_select_comp, urwid.Divider(), m_exit])

        view = urwid.Filler(
            urwid.Pile(
                [
                    hpad,
                    urwid.Text(self._get_version(), align=urwid.CENTER),
                    urwid.Padding(
                        urwid.LineBox(menu, "Main Menu", title_align="left"),
                        width=("relative", 80),
                        align=urwid.CENTER,
                    ),
                ]
            ),
            "middle",
        )
        return view

    def _on_exit(self, btn):
        self.finish(None)

    def _run_screen(self, screen_name, btn=None):
        # We want to import screen class as late as possible to avoid delays
        if screen_name == "welcome":
            from compman.ui.welcome import WelcomeScreen

            screen = WelcomeScreen(self.container)
        elif screen_name == "select":
            from compman.ui.selectcomp import SelectCompetitionScreen

            screen = SelectCompetitionScreen(self.container)
        else:
            assert screen_name == "details"
            from compman.ui.compdetails import CompetitionDetailsScreen

            screen = CompetitionDetailsScreen(self.container)
        screen.show()

    def _get_version(self) -> str:
        return f"Version {compman.__version__}"



"""Welcome screen"""
import urwid

import compman
from compman import storage
from compman.ui import widget
from compman.ui.activity import Activity


class WelcomeScreen(Activity):
    def create_view(self):
        btxt = urwid.BigText("Compman", urwid.font.Thin6x6Font())
        hpad = urwid.Padding(urwid.AttrMap(btxt, "screen header"), "center", "clip")

        intro_text = [
            "Welcome to ",
            ("screen header", "Competition Manager"),
            f" version {compman.__version__}",
            "! ",
            "This app allows you to keep your contest files, ",
            "such as airspace or turnpoint, up to date. ",
            "To begin, pick your competition.",
        ]

        add_comp_button = widget.CMButton("Set Up a Competition")
        self.connect_async(add_comp_button, "click", self._pick_competition)

        intro = urwid.Padding(
            urwid.Pile(
                [
                    urwid.Divider(),
                    urwid.Text(intro_text),
                    urwid.Divider(),
                    urwid.GridFlow([add_comp_button], 24, 2, 1, "left"),
                    urwid.Divider(),
                ]
            ),
            left=1,
            right=1,
        )

        view = urwid.Filler(
            urwid.Pile(
                [
                    hpad,
                    urwid.Padding(
                        urwid.LineBox(intro, "Welcome!", title_align="left"),
                        width=("relative", 80),
                        align="center",
                    ),
                ]
            ),
            "middle",
        )

        return view

    async def _pick_competition(self):
        from compman.ui.soaringspot import SoaringSpotPickerScreen

        comp = await self.run_activity(SoaringSpotPickerScreen(self.container))
        if comp is None:
            return

        storage.get_settings().current_competition_id = comp.id
        storage.save_settings()
        self.finish(comp)

        from compman.ui.compdetails import CompetitionDetailsScreen

        details = CompetitionDetailsScreen(self.container)
        details.show()



from typing import Optional
import asyncio
import logging

import urwid


log = logging.getLogger("compman")


class CMButton(urwid.Button):
    def __init__(self, text):
        super().__init__(text)
        txt = urwid.SelectableIcon(text)
        txt.align = "center"
        btn = urwid.AttrWrap(txt, "btn normal", "btn focus")
        self._w = btn


class CMSelectableListItem(urwid.Button):
    def __init__(self, text):
        super().__init__(text)
        btn = urwid.SelectableIcon(text)
        wdg = urwid.AttrMap(btn, "li normal", "li focus")
        self._w = wdg

    def _btnclicked(self, btn):
        self._emit("selected")


class CMScreenPadding(urwid.Padding):
    def __init__(self, w):
        super().__init__(w, width=(urwid.RELATIVE, 92), align=urwid.CENTER)


class CMScreenHeader(urwid.WidgetWrap):
    def __init__(self, text: str) -> None:
        orig = urwid.AttrMap(
            urwid.Pile(
                [
                    urwid.Divider(),
                    CMScreenPadding(urwid.Text(text)),
                    urwid.Divider("─"),
                    urwid.Divider(),
                ]
            ),
            "screen header",
        )
        super().__init__(orig)


class CMFlashMessage(urwid.WidgetWrap):
    _flashtask: Optional[asyncio.Task]

    def __init__(self, activity) -> None:
        self.text = urwid.Text("")
        self._activity = activity
        self._flashtask = None
        super().__init__(self.text)

    def flash(self, markup: str) -> None:
        if self._flashtask and not self._flashtask.done():
            self._flashtask.cancel()
        self._flashtask = self._activity.async_task(self._flash_status(markup))

    async def _flash_status(self, markup: str):
        self.text.set_text(markup)
        await asyncio.sleep(3.0)
        self.text.set_text("")


class CMGlobalCommands(urwid.WidgetWrap):
    def __init__(self, widget, activity):
        super().__init__(widget)
        self.activity = activity

    def keypress(self, size, key):
        if key == "esc":
            self.activity.finish(None)
            return
        return self._w.keypress(size, key)



import pytest

from compman import main


@pytest.mark.asyncio
async def test_main(mocker, activity_testbed) -> None:
    # GIVEN
    mocker.patch("compman.main.xcsoar")
    mocker.patch("compman.main.storage")
    mocker.patch("compman.main.logging")
    mocker.patch("compman.main.urwid")
    mocker.patch("compman.ui.mainmenu.MainMenuScreen")

    main.run([])


def test_debounce_esc() -> None:
    assert main.debounce_esc(["down"], []) == ["down"]
    assert main.debounce_esc(["down", "down"], []) == ["down", "down"]
    assert main.debounce_esc(["esc"], []) == ["esc"]
    assert main.debounce_esc(["esc", "esc"], []) == ["esc"]
    assert main.debounce_esc(["esc", "esc", "esc"], []) == ["esc", "esc"]
    assert main.debounce_esc(["esc", "down", "esc"], []) == ["esc", "down", "esc"]



import pytest

from compman import storage
from compman.ui.welcome import WelcomeScreen


@pytest.mark.asyncio
async def test_welcomescreen_view(activity_testbed) -> None:
    # View is opened with the welcome message
    async with activity_testbed.shown(WelcomeScreen):
        # WHEN
        contents = activity_testbed.render()

        # THEN
        assert "Welcome to Competition Manager" in contents
        assert "Set Up a Competition" in contents


@pytest.mark.asyncio
async def test_welcomescreen_setup_comp(
    storage_dir, soaringspot, activity_testbed
) -> None:
    # User picks a competition and end up in competition details screen
    # GIVEN
    sspicker_screen = activity_testbed.mock(
        "compman.ui.soaringspot.SoaringSpotPickerScreen"
    )
    sspicker_screen.result = storage.StoredCompetition(
        id="test", title="Test", soaringspot_url="http://test"
    )
    details_screen = activity_testbed.mock(
        "compman.ui.compdetails.CompetitionDetailsScreen"
    )

    async with activity_testbed.shown(WelcomeScreen):
        # WHEN
        await activity_testbed.keypress("enter")

        # THEN
        assert sspicker_screen.shown
        assert details_screen.shown
        assert storage.get_settings().current_competition_id == "test"


@pytest.mark.asyncio
async def test_welcomescreen_setup_cancelled(soaringspot, activity_testbed) -> None:
    # User chooses to set up the competition but presses escape and gets back
    # to welcome screen.
    sspicker_screen = activity_testbed.mock(
        "compman.ui.soaringspot.SoaringSpotPickerScreen"
    )
    details_screen = activity_testbed.mock(
        "compman.ui.compdetails.CompetitionDetailsScreen"
    )

    async with activity_testbed.shown(WelcomeScreen):
        # WHEN
        await activity_testbed.keypress("enter")
        await activity_testbed.keypress("esc")

        # THEN
        assert sspicker_screen.shown
        assert not details_screen.shown

        contents = activity_testbed.render()
        assert "Welcome to Competition Manager" in contents
        assert "Set Up a Competition" in contents



import pytest

from compman import storage
from compman.ui.mainmenu import MainMenuScreen


@pytest.mark.asyncio
async def test_mainmenu_initial_welcome(storage_dir, activity_testbed) -> None:
    # View is opened with the welcome message
    welcome_screen = activity_testbed.mock("compman.ui.welcome.WelcomeScreen")
    async with activity_testbed.shown(MainMenuScreen):
        # THEN
        assert welcome_screen.shown


@pytest.mark.asyncio
async def test_mainmenu_details(storage_dir, activity_testbed) -> None:
    # GIVEN
    welcome_screen = activity_testbed.mock("compman.ui.welcome.WelcomeScreen")
    details_screen = activity_testbed.mock(
        "compman.ui.compdetails.CompetitionDetailsScreen"
    )
    async with activity_testbed.shown(MainMenuScreen):
        # WHEN
        assert welcome_screen.shown
        # Go back to main menu
        await activity_testbed.keypress("esc")

        contents = activity_testbed.render()
        assert "Current Competition" in contents

        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "Current Competition"
        await activity_testbed.keypress("enter")

        # THEN
        assert details_screen.shown


@pytest.mark.asyncio
async def test_mainmenu_select_comp(storage_dir, activity_testbed) -> None:
    welcome_screen = activity_testbed.mock("compman.ui.welcome.WelcomeScreen")
    select_screen = activity_testbed.mock(
        "compman.ui.selectcomp.SelectCompetitionScreen"
    )
    async with activity_testbed.shown(MainMenuScreen):
        # WHEN
        assert welcome_screen.shown
        # Go back to main menu
        await activity_testbed.keypress("esc")

        contents = activity_testbed.render()
        assert "Select Competition" in contents

        await activity_testbed.keypress("down")
        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "Select Competition"
        await activity_testbed.keypress("enter")

        # THEN
        assert select_screen.shown


@pytest.mark.asyncio
async def test_mainmenu_exit(storage_dir, activity_testbed) -> None:
    # GIVEN
    welcome_screen = activity_testbed.mock("compman.ui.welcome.WelcomeScreen")

    res = {"exited": False}

    def onexit():
        res["exited"] = True

    async with activity_testbed.shown(MainMenuScreen) as activity:
        activity.on_exit(onexit)

        # WHEN
        assert welcome_screen.shown
        # Go back to main menu
        await activity_testbed.keypress("esc")

        # THEN
        contents = activity_testbed.render()
        assert "Exit" in contents

        await activity_testbed.keypress("down")
        await activity_testbed.keypress("down")
        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "Exit"
        await activity_testbed.keypress("enter")

        # THEN
        assert res["exited"] == True


@pytest.mark.asyncio
async def test_mainmenu_details_if_comp_selected(storage_dir, activity_testbed) -> None:
    # GIVEN
    welcome_screen = activity_testbed.mock("compman.ui.welcome.WelcomeScreen")
    details_screen = activity_testbed.mock(
        "compman.ui.compdetails.CompetitionDetailsScreen"
    )
    storage.get_settings().current_competition_id = "test"
    storage.save_settings()
    async with activity_testbed.shown(MainMenuScreen):
        # THEN
        assert not welcome_screen.shown
        assert details_screen.shown



from typing import Optional, Type, List
import asyncio
from contextlib import asynccontextmanager

import urwid

from compman.ui.activity import Activity


class ActivityStub(Activity):
    _result: Optional[object] = None

    def __init__(
        self, container, monitor: "ActivityStubMonitor", result: Optional[object]
    ) -> None:
        super().__init__(container)
        self._monitor = monitor
        self._result = result
        self._monitor.created = True

    def create_view(self) -> None:
        return urwid.SolidFill("T")

    def show(self) -> None:
        super().show()
        self._monitor.shown = True
        if self._result is not None:
            self.finish(self._result)


class ActivityStubMonitor:
    def __init__(self, result: object = None) -> None:
        self.result = result
        self.created = False
        self.shown = False

    def create(self, container):
        return ActivityStub(container, self, self.result)


class ActivityTestbed:
    def __init__(self, mocker) -> None:
        self.mocker = mocker
        self.container = urwid.WidgetPlaceholder(urwid.SolidFill("T"))
        self.activity: Optional[Activity] = None

    @asynccontextmanager
    async def shown(self, act_factory: Type[Activity]):
        self.activity = act_factory(self.container)
        self.activity.show()

        yield self.activity

        await self.gather_tasks()
        if not self.activity.is_finished():
            self.activity.finish(None)
        self.activity = None

    async def gather_tasks(self):
        await asyncio.wait_for(asyncio.gather(*self.activity._tasks), 1)

    async def response(self):
        return await self.activity.response

    def mock(self, spec, result: object = None) -> ActivityStubMonitor:
        monitor = ActivityStubMonitor(result)
        self.mocker.patch(spec, monitor.create)
        return monitor

    def render(self, size=(60, 30)):
        canvas = self.container.render(size)
        contents = [t.decode("utf-8") for t in canvas.text]
        return "\n".join(contents)

    def get_focus_widgets(self) -> List[urwid.Widget]:
        container = self._find_container_widget(self.container)
        return container.get_focus_widgets()

    async def keypress(self, *keys: str) -> None:
        for key in keys:
            res = self.container.keypress((60, 30), key)
            assert res is None, "Keypress is not processed"
            await asyncio.sleep(0)

    def _find_container_widget(self, w: urwid.Widget) -> urwid.WidgetContainerMixin:
        if isinstance(w, urwid.WidgetContainerMixin):
            return w
        if isinstance(w, urwid.WidgetDecoration):
            return self._find_container_widget(w.original_widget)
        if isinstance(w, urwid.WidgetWrap):
            return self._find_container_widget(w._w)
        raise RuntimeError(f"Unknown widget type: {w}")



from typing import List, Union, Tuple, cast
import logging
import asyncio

import urwid

from compman import storage
from compman import soaringspot
from compman import xcsoar
from compman.ui import widget
from compman.ui.activity import Activity

nothing = urwid.Pile([])

log = logging.getLogger("compman")

UrwidMarkup = List[Union[str, Tuple[str, str]]]


class CompetitionDetailsScreen(Activity):
    def show(self):
        cid = storage.get_settings().current_competition_id
        self.competition = storage.load_competition(cid)
        self.airspaces = storage.get_airspace_files(cid)
        self.waypoints = storage.get_waypoint_files(cid)
        self.profiles = xcsoar.list_xcsoar_profiles()

        super().show()
        self._flashtask = None
        self.async_task(self._update_competition_files())

    def create_view(self):
        p2 = lambda w: urwid.Padding(w, left=2)

        self.airspace_group = []
        self.airspace_pile = urwid.Pile(
            [
                self._make_file_radio(
                    sf,
                    self.airspace_group,
                    sf.name == self.competition.airspace,
                    self._on_airspace_changed,
                )
                for sf in self.airspaces
            ]
        )

        self.waypoint_group = []
        self.waypoint_pile = urwid.Pile(
            [
                self._make_file_radio(
                    sf,
                    self.waypoint_group,
                    sf.name == self.competition.waypoints,
                    self._on_waypoint_changed,
                )
                for sf in self.waypoints
            ]
        )

        self.profile_pile = urwid.Pile(
            [self._make_profile_checkbox(prf) for prf in self.profiles]
        )

        self.download_status = urwid.Text("")

        form = urwid.Pile(
            [
                self.download_status,
                urwid.Divider(),
                urwid.Text("Airspace files"),
                p2(self.airspace_pile),
                urwid.Divider(),
                urwid.Text("Waypoint files"),
                p2(self.waypoint_pile),
                urwid.Divider(),
                urwid.Text("XCSoar profiles"),
                p2(self.profile_pile),
                urwid.Divider(),
                self._create_buttons(),
            ]
        )

        filler = urwid.Filler(form, valign=urwid.TOP)

        self.status = widget.CMFlashMessage(self)

        return urwid.Frame(
            widget.CMScreenPadding(filler),
            header=widget.CMScreenHeader(self.competition.title),
            footer=widget.CMScreenPadding(self.status),
        )

    def _create_buttons(self):
        activate_btn = widget.CMButton("Activate")
        urwid.connect_signal(activate_btn, "click", self._on_activate)

        exit_btn = widget.CMButton("Main Menu")
        urwid.connect_signal(exit_btn, "click", self._on_exit)
        return urwid.GridFlow([activate_btn, exit_btn], 22, 2, 1, "left")

    def _on_airspace_changed(self, ev, new_state, selected):
        if not new_state:
            return
        self.competition.airspace = selected
        storage.save_competition(self.competition)
        self._udpate_xcsoar_config(self.competition)
        self.status.flash(f"Airspace changed to: {selected}")

    def _on_waypoint_changed(self, ev, new_state, selected):
        if not new_state:
            return
        self.competition.waypoints = selected
        storage.save_competition(self.competition)
        self._udpate_xcsoar_config(self.competition)
        self.status.flash(f"Waypoint changed to: {selected}")

    def _on_profile_changed(self, ev, selected, profile: str) -> None:
        if selected:
            self.competition.add_profile(profile)
        else:
            self.competition.remove_profile(profile)
        storage.save_competition(self.competition)

    def _on_activate(self, btn) -> None:
        if not self.competition.profiles:
            self.status.flash(
                ("error message", "Please select at least one XCSoar profile")
            )
            return

        profiles = ", ".join(self.competition.profiles)
        self._udpate_xcsoar_config(self.competition)
        self.status.flash(("success message", f"XCSoar profiles updated: {profiles}"))

    def _on_exit(self, btn):
        self.finish(None)

    def _udpate_xcsoar_config(self, comp: storage.StoredCompetition) -> None:
        for prf in self.competition.profiles:
            xcprofile = xcsoar.get_xcsoar_profile(prf)
            if comp.airspace:
                xcprofile.set_airspace(
                    storage.get_full_file_path(comp.id, comp.airspace)
                )
            if comp.waypoints:
                xcprofile.set_waypoint(
                    storage.get_full_file_path(comp.id, comp.waypoints)
                )
            xcprofile.save()

    async def _update_competition_files(self):
        compurl = self.competition.soaringspot_url
        if compurl is None:
            return

        self.download_status.set_text(("progress", "Refreshing file list..."))
        downloads = await soaringspot.fetch_downloads(compurl)
        new_airspaces, new_waypoints = self._detect_new_files(downloads)

        if not new_airspaces and not new_waypoints:
            self.download_status.set_text(("remark", "No updates"))
            return

        self.download_status.set_text(
            ("success message", "New contest files detected!")
        )

        tasks = []
        tasks.extend(
            self._download_new_files(
                new_airspaces,
                self.airspace_pile,
                self.airspace_group,
                self._on_airspace_changed,
            )
        )
        tasks.extend(
            self._download_new_files(
                new_waypoints,
                self.waypoint_pile,
                self.waypoint_group,
                self._on_waypoint_changed,
            )
        )
        await asyncio.gather(*tasks)

    def _detect_new_files(self, downloads):
        DFT = soaringspot.DownloadableFileType
        airspace_files = {f.name for f in self.airspaces}
        dl_airspaces = {d.filename for d in downloads if d.kind == DFT.AIRSPACE}
        new_airspaces = dl_airspaces - airspace_files

        waypoint_files = {f.name for f in self.waypoints}
        dl_waypoints = {d.filename for d in downloads if d.kind == DFT.WAYPOINT}
        new_waypoints = dl_waypoints - waypoint_files

        dl_idx = {d.filename: d for d in downloads}
        return ([dl_idx[f] for f in new_airspaces], [dl_idx[f] for f in new_waypoints])

    def _download_new_files(self, new_files, checkbox_pile, group, select_handler):
        tasks = []
        for sspotfile in new_files:
            sf = storage.StoredFile(name=sspotfile.filename, size=None)
            radio = self._make_file_radio(sf, group, False, select_handler)
            checkbox_pile.contents.insert(0, (radio, ("pack", None)))
            tasks.append(self._download_file(sf, sspotfile.href, radio))

        return tasks

    async def _download_file(
        self, sf: storage.StoredFile, url: str, radio: urwid.RadioButton
    ):
        orig_radio = radio.original_widget
        orig_radio.set_label(self._make_label(sf, [("progress", "Downloading...")]))
        dlcontents = await soaringspot.fetch_file(url)
        stored = storage.store_file(self.competition.id, sf.name, dlcontents)
        orig_radio.set_label(self._make_label(stored, [("success banner", " New! ")]))

    def _make_file_radio(
        self, sf: storage.StoredFile, group, selected: bool, select_handler
    ) -> urwid.RadioButton:
        label = f"{sf.name} ({sf.format_size()})"
        radio = urwid.RadioButton(group, label, selected)
        urwid.connect_signal(radio, "change", select_handler, sf.name)
        return urwid.AttrMap(radio, "li normal", "li focus")

    def _make_label(
        self, sf: storage.StoredFile, extra: UrwidMarkup = None
    ) -> UrwidMarkup:
        label = f"{sf.name} ({sf.format_size()})"
        markup = cast(UrwidMarkup, [label, " "])
        markup.extend(extra or [])
        return markup

    def _make_profile_checkbox(self, profile: str) -> urwid.CheckBox:
        cb = urwid.CheckBox(profile, profile in self.competition.profiles)
        urwid.connect_signal(cb, "change", self._on_profile_changed, profile)
        return urwid.AttrMap(cb, "li normal", "li focus")



import asyncio

import pytest

from compman import storage
from compman import xcsoar
from compman.soaringspot import SoaringSpotDownloadableFile, DownloadableFileType
from compman.ui.compdetails import CompetitionDetailsScreen


@pytest.mark.asyncio
async def test_compdetails_view(
    storage_dir, soaringspot, xcsoar_dir, activity_testbed
) -> None:
    # GIVEN
    _setup_test_comp()
    async with activity_testbed.shown(CompetitionDetailsScreen):
        assert "Test Competition" in activity_testbed.render()
        await asyncio.sleep(0)
        assert "Refreshing file list..." in activity_testbed.render()
        await asyncio.sleep(0)
        assert "No updates" in activity_testbed.render()


@pytest.mark.asyncio
async def test_compdetails_download(
    storage_dir, soaringspot, xcsoar_dir, activity_testbed
) -> None:
    # GIVEN
    comp = _setup_test_comp()
    soaringspot.files = [
        SoaringSpotDownloadableFile(
            "airspace.txt",
            href=f"{comp.soaringspot_url}/airspace.txt",
            kind=DownloadableFileType.AIRSPACE,
        )
    ]
    soaringspot.file_contents = b"Hello World!"

    # WHEN
    async with activity_testbed.shown(CompetitionDetailsScreen):
        await asyncio.sleep(0)
        assert "Refreshing file list..." in activity_testbed.render()
        await asyncio.sleep(0)
        assert "New contest files detected!" in activity_testbed.render()
        assert "airspace.txt" in activity_testbed.render()
        await asyncio.sleep(0)

        assert "Downloading..." in activity_testbed.render()
        await asyncio.sleep(0)
        assert "New!" in activity_testbed.render()
        assert "airspace.txt (12.0B)" in activity_testbed.render()
        await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_compdetails_activate(
    storage_dir, soaringspot, activity_testbed, xcsoar_dir, async_sleep
) -> None:
    # GIVEN
    comp = _setup_test_comp()
    soaringspot.files = [
        SoaringSpotDownloadableFile(
            "airspace.txt",
            href=f"{comp.soaringspot_url}/airspace.txt",
            kind=DownloadableFileType.AIRSPACE,
        ),
        SoaringSpotDownloadableFile(
            "waypoints.cup",
            href=f"{comp.soaringspot_url}/waypoints.cup",
            kind=DownloadableFileType.WAYPOINT,
        ),
    ]
    soaringspot.file_contents = b"Hello World!"

    async with activity_testbed.shown(CompetitionDetailsScreen):
        # Let everything to be downloaded
        await activity_testbed.gather_tasks()

        assert "New contest files detected!" in activity_testbed.render()

        await activity_testbed.keypress("enter", "down")
        focused = activity_testbed.get_focus_widgets()[-1]
        assert focused.get_label() == "Activate"
        await activity_testbed.keypress("enter")

        # THEN
        assert "XCSoar profiles updated" in activity_testbed.render()


@pytest.mark.asyncio
async def test_compdetails_select_files(
    storage_dir, soaringspot, activity_testbed, xcsoar_dir, async_sleep
) -> None:
    # GIVEN
    comp = _setup_test_comp()
    soaringspot.files = [
        SoaringSpotDownloadableFile(
            "airspace.txt",
            href=f"{comp.soaringspot_url}/airspace.txt",
            kind=DownloadableFileType.AIRSPACE,
        ),
        SoaringSpotDownloadableFile(
            "waypoints.cup",
            href=f"{comp.soaringspot_url}/waypoints.cup",
            kind=DownloadableFileType.WAYPOINT,
        ),
    ]
    soaringspot.file_contents = b"Hello World!"

    # WHEN
    async with activity_testbed.shown(CompetitionDetailsScreen):
        # Let everything to be downloaded
        await activity_testbed.gather_tasks()

        assert "New contest files detected!" in activity_testbed.render()

        # Now select profile, airspace and waypoint files
        await activity_testbed.keypress("enter", "up", "enter", "up", "enter")

        # THEN
        content = activity_testbed.render()
        assert "(X) airspace.txt" in content
        assert "(X) waypoints.cup" in content
        assert "Airspace changed to: airspace.txt" in content
        assert "[X] openvario.prf" in content

    with open(xcsoar.get_xcsoar_profile_filename("openvario.prf"), "r") as f:
        profile_contents = f.read()

    assert "waypoints.cup" in profile_contents
    assert "airspace.txt" in profile_contents


def _setup_test_comp() -> storage.StoredCompetition:
    comp = storage.StoredCompetition(
        id="test",
        title="Test Competition",
        soaringspot_url="https://soaringspot.local/test",
    )
    storage.save_competition(comp, set_current=True)
    return comp



import pytest

from compman import storage
from compman.ui.selectcomp import SelectCompetitionScreen


@pytest.mark.asyncio
async def test_selectcomp_view_initial(storage_dir, activity_testbed) -> None:
    async with activity_testbed.shown(SelectCompetitionScreen):
        content = activity_testbed.render()
        assert "Switch competition" in content
        focus = activity_testbed.get_focus_widgets()[-1]
        assert "New competition" in focus.get_label()


@pytest.mark.asyncio
async def test_selectcomp_view_existing(storage_dir, activity_testbed) -> None:
    # GIVEN

    details_screen = activity_testbed.mock(
        "compman.ui.compdetails.CompetitionDetailsScreen"
    )

    storage.save_competition(storage.StoredCompetition("one", title="One"))
    storage.save_competition(storage.StoredCompetition("two", title="Two"))

    # WHEN
    async with activity_testbed.shown(SelectCompetitionScreen):
        content = activity_testbed.render()

        assert "One" in content
        assert "Two" in content
        focus = activity_testbed.get_focus_widgets()[-1]
        assert focus.get_label() == "One"

        await activity_testbed.keypress("enter")

        # THEN
        assert details_screen.shown

        settings = storage.get_settings()
        assert settings.current_competition_id == "one"


@pytest.mark.asyncio
async def test_selectcomp_new_comp_cancel(storage_dir, activity_testbed) -> None:
    sspicker_screen = activity_testbed.mock(
        "compman.ui.soaringspot.SoaringSpotPickerScreen"
    )

    async with activity_testbed.shown(SelectCompetitionScreen):
        content = activity_testbed.render()
        assert "Switch competition" in content
        focus = activity_testbed.get_focus_widgets()[-1]
        assert "New competition" in focus.get_label()
        await activity_testbed.keypress("enter")
        await activity_testbed.keypress("esc")

        assert sspicker_screen.shown
        content = activity_testbed.render()
        assert "Switch competition" in content


@pytest.mark.asyncio
async def test_selectcomp_new_comp_select(storage_dir, activity_testbed) -> None:
    sspicker_screen = activity_testbed.mock(
        "compman.ui.soaringspot.SoaringSpotPickerScreen"
    )
    sspicker_screen.result = storage.StoredCompetition("one", title="One")

    details_screen = activity_testbed.mock(
        "compman.ui.compdetails.CompetitionDetailsScreen"
    )

    async with activity_testbed.shown(SelectCompetitionScreen):
        focus = activity_testbed.get_focus_widgets()[-1]
        assert "New competition" in focus.get_label()
        await activity_testbed.keypress("enter")

        assert sspicker_screen.shown
        assert details_screen.shown



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="openvario-compman",
    version="0.4",
    description="Competition Manager for OpenVario",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kedder/openvario-compman",
    author="Andrey Lebedev",
    author_email="andrey.lebedev@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="openvario tui competition igc xcsoar",
    package_dir={"": "src"},
    packages=find_packages(where="src"),  # Required
    python_requires=">=3.7, <4",
    install_requires=["urwid", "aiohttp", "lxml", "openvario-shell"],
    extras_require={
        "dev": ["black", "mypy"],
        "test": ["pytest", "pytest-coverage", "pytest-asyncio", "pytest-mock"],
    },
    package_data={},
    data_files=[],
    entry_points={
        "console_scripts": ["compman=compman.main:main"],
        "ovshell.extensions": ["compman=compman.ovshell:extension"],
    },
    project_urls={},
)



from typing import List

import urwid

from compman import soaringspot
from compman import storage
from compman.ui.activity import Activity
from compman.ui import widget


class SoaringSpotPicker(urwid.ListBox):
    signals = ["select", "focus"]

    def __init__(self, activity: Activity) -> None:
        self._items = urwid.SimpleListWalker([])
        super().__init__(self._items)
        self.activity = activity
        activity.async_task(self._download_competitions())

    def set_competitions(
        self, competitions: List[soaringspot.SoaringSpotContest]
    ) -> None:
        del self._items[:]
        if not competitions:
            return

        self.competitions = competitions
        for comp in competitions:
            btn = widget.CMSelectableListItem(comp.title)
            urwid.connect_signal(btn, "click", self._on_competition_selected, comp)

            # txt = urwid.Padding(
            #     urwid.AttrMap(urwid.Text(comp.description), "text"), left=2
            # )
            self._items.extend([btn])
        urwid.connect_signal(self._items, "modified", self._on_focus_changed)
        self._emit("focus", competitions[0])

    async def _download_competitions(self):
        del self._items[:]
        self._items.append(urwid.Text(("progress", "Downloading...")))
        comps = await soaringspot.fetch_competitions()
        self.set_competitions(comps)

    def _on_competition_selected(
        self, btn: urwid.Widget, sscomp: soaringspot.SoaringSpotContest
    ) -> None:
        comp = storage.load_competition(sscomp.id)
        if comp is not None:
            comp.title = sscomp.title
            comp.soaringspot_url = sscomp.href
        else:
            comp = storage.StoredCompetition(
                id=sscomp.id, title=sscomp.title, soaringspot_url=sscomp.href
            )
        storage.save_competition(comp)
        self._emit("select", comp)

    def _on_focus_changed(self):
        focused, idx = self.get_focus()
        selected = self.competitions[idx]
        self._emit("focus", selected)


class SoaringSpotPickerScreen(Activity):
    def create_view(self):
        mainview = SoaringSpotPicker(self)
        urwid.connect_signal(mainview, "select", self._on_competition_selected)
        urwid.connect_signal(mainview, "focus", self._on_comp_focused)
        self.footer = urwid.Text("")

        return urwid.Frame(
            widget.CMScreenPadding(mainview),
            header=widget.CMScreenHeader("Pick a new competition"),
            footer=widget.CMScreenPadding(self.footer),
        )

    def _on_competition_selected(self, ev, comp):
        self.finish(comp)

    def _on_comp_focused(self, ev, comp):
        self.footer.set_text(comp.description)



from typing import Set
import asyncio
import logging

import urwid

from compman.ui import widget

log = logging.getLogger("compman")


class Activity:
    _previous_widget: urwid.Widget = None
    _tasks: Set[asyncio.Task]

    def __init__(self, container):
        self.container = container
        self.response = asyncio.Future()

    def create_view(self) -> urwid.Widget:
        raise NotImplementedError()

    def show(self) -> None:
        self._tasks = set()
        self._previous_widget = self.container.original_widget
        self.container.original_widget = widget.CMGlobalCommands(
            self.create_view(), self
        )

    def finish(self, result):
        if self._previous_widget is None:
            raise RuntimeError("Activity is not started or already finished")
        # Cancel any outstanding tasks on a next event loop iteration to avoid
        # cancelling the task that calls finish() right now.
        asyncio.get_running_loop().call_soon(self._cancel_outstanding_tasks)

        # Restore the prvious UI
        self.container.original_widget = self._previous_widget
        self._previous_widget = None

        self.response.set_result(result)

    def is_finished(self) -> bool:
        return self._previous_widget is None

    def connect_async(self, widget, signal, handler_coro, user_args=None):
        urwid.connect_signal(
            widget,
            signal,
            self._async_handler,
            user_args=[handler_coro, user_args or []],
        )

    async def run_activity(self, act: "Activity"):
        log.debug(f"Running activity {act.__class__}")
        act.show()
        resp = await act.response
        log.debug(f"Activity {act.__class__} completed with {resp}")
        return resp

    def async_task(self, coro):
        task = asyncio.create_task(coro)
        log.debug(f"Started {task}")
        task.add_done_callback(self._task_done)
        self._tasks.add(task)
        return task

    def _async_handler(self, handler_coro, args, widget):
        coro = handler_coro(*args)
        return self.async_task(coro)

    def _cancel_outstanding_tasks(self):
        # Cancel any outstanding tasks
        notdone = [t for t in self._tasks if not t.done()]
        if notdone:
            log.info(
                f"Cancelling {len(notdone)} outstanding tasks "
                f"for {self.__class__.__name__}"
            )
        for t in notdone:
            t.cancel()

    def _task_done(self, task):
        self._tasks.remove(task)
        if task.cancelled():
            log.debug(f"Cancelled: {task}")
            return
        exc = task.exception()
        if exc is not None:
            log.error(f"Failed: {task}", exc_info=exc)
        else:
            log.debug(f"Completed: {task}")



import asyncio

import pytest

from ovshell.testing import OpenVarioShellStub
from compman.ovshell import extension, CompmanShellApp, CompmanShellActivity


@pytest.fixture
def ovshell(tmp_path: str) -> OpenVarioShellStub:
    return OpenVarioShellStub(tmp_path)


@pytest.fixture
def compman_app(
    ovshell: OpenVarioShellStub, tmp_path: str, monkeypatch, xcsoar_dir: str
) -> CompmanShellApp:
    app = CompmanShellApp(ovshell)
    monkeypatch.setenv("COMPMAN_XCSOARDIR", str(xcsoar_dir))
    monkeypatch.setenv("COMPMAN_DATADIR", str(tmp_path))
    return app


def test_extension(ovshell: OpenVarioShellStub) -> None:
    # WHEN
    ext = extension("compman", ovshell)

    apps = ext.list_apps()
    assert len(apps) == 1


def test_launch_app(ovshell: OpenVarioShellStub, compman_app: CompmanShellApp) -> None:
    # WHEN
    compman_app.launch()

    # THEN
    act = ovshell.screen.stub_top_activity()
    assert isinstance(act, CompmanShellActivity)


def test_ovshell_activity(
    ovshell: OpenVarioShellStub, compman_app: CompmanShellApp
) -> None:
    # WHEN
    compman_app.launch()
    act = ovshell.screen.stub_top_activity()
    assert act is not None
    widget = act.create()

    # THEN
    canvas = widget.render((60, 30))
    contents = "\n".join([t.decode("utf-8") for t in canvas.text])
    assert "Welcome to Competition Manager" in contents


@pytest.mark.asyncio
async def test_ovshell_exit_app(
    ovshell: OpenVarioShellStub, compman_app: CompmanShellApp
) -> None:
    # WHEN
    compman_app.launch()
    act = ovshell.screen.stub_top_activity()
    assert act is not None
    widget = act.create()

    widget.keypress((60, 30), "esc")
    await asyncio.sleep(0)

    canvas = widget.render((60, 30))
    contents = "\n".join([t.decode("utf-8") for t in canvas.text])
    assert "Main Menu" in contents

    widget.keypress((60, 30), "esc")
    await asyncio.sleep(0)

    # THEN
    assert ovshell.screen.stub_top_activity() is None
